#!/usr/bin/pwsh

# Settings

nv JAR_PATTERN       "*upgrade-tool-*.jar"
nv JAVA_BIN          "java"
nv REQUIRED_JAVA     "1.8.0"

# Global variables

nv -Option "AllScope" JAVA_PATH                $null
nv -Option "AllScope" ACTION                   $null
nv -Option "AllScope" MULE_HOME_OUTDATED       $null
nv -Option "AllScope" MULE_HOME_UPDATED        $null
nv -Option "AllScope" OUTDATED_JAR             $null
nv -Option "AllScope" UPDATED_JAR              $null
nv -Option "AllScope" OLD_AGENT_VERSION        $null
nv -Option "AllScope" NEW_AGENT_VERSION        $null

# Strings

$Msg = @(0..20)
$Msg[0] = "ERROR: {0} version is {1}, minimum required is {2}, exiting"
$Msg[1] = "ERROR: Java binary not found neither in JAVA nor in JAVA_HOME, exiting"
$Msg[2] = "ERROR: Please use absolute paths for MULE_HOMEs, exiting" # <--- Not used in Windows
$Msg[3] = "INFO: Detected outdated version"
$Msg[4] = "INFO: Upgrading from {0} to {1}"
$Msg[5] = "INFO: Upgrade terminated with status {0}"
$Msg[6] = "INFO: Skipping update, already at latest version ({0})"
$Msg[7] = "ERROR: MULE_HOME seems invalid, exiting"
$Msg[8] = "`nChecking Mule Agent"
$Msg[9] = "`nMule Runtime Upgrade Tool"
$Msg[10] = "INFO: Mule Agent is not present in the current setup, but it seems it was installed before, copying it to {0}"
$Msg[11] = "INFO: Mule Agent not installed, skipping setup"
$Msg[12] = "INFO: No update available, skipping update"

function New-Separator ( $Title ) {

  # Colorize
  [Console]::ForegroundColor = [System.ConsoleColor]::Green

  # Print
  "{0}" -f $Title
  "{0}`n" -f ( "-" * $Title.Length )

  # Reset color
  [Console]::ResetColor()
}

function Compare-Versions {

  # Set variables
  begin {
    $Err  = $false
    $NAME   = $Params[0]
    $CUR  = $Params[1].Split('.')
    $REQ  = $Params[2].Split('.')
  }

  # Compare versions
  process {
    if ( $CUR[0] -lt $REQ[0] ) {
      $Err = $true
    }
    elseif ( $CUR[0] -eq $REQ[0] ) {
      if ( $CUR[1] -lt $REQ[1] ) {
        $Err = $true
      }
      elseif ( $CUR[1] -eq $REQ[1] ) {
        if ( $CUR[2] -lt $REQ[2] ) {
          $Err = $true
        }
      }
    }
  }

  # Report error
  end {
    if ( $Err ) {
      return $true
    }
    else {
      return $false
    }
  }
}

function Search-in-Path {
  try {
    $p = Get-Command -Type "Application" $JAVA_BIN 2> $null
    if ( $? ) {
      $JAVA_PATH = $p[0].Path
      return $true
    }
    else {
      throw
    }
  }
  catch {
    return $false
  }
}

function Search-in-Env {
  try {
    if ( $env:JAVA_HOME ) {
      $JAVA_PATH = $env:JAVA_HOME + "/bin/" + $JAVA_BIN

      if ( Test-Path -Path "$JAVA_PATH" -PathType Leaf ) {
        return $true
      }

      else {
        throw
      }
    }
  }
  catch {
    return $false
  }
}

function Search-Java-Bin {
  if ( Search-in-Path ) {
    return $true
  }
  elseif ( Search-in-Env ) {
    return $true
  }
  else {
    [Console]::Error.WriteLine( $Msg[1] )
    exit 1
  }
}

function Check-JVM-Version {
  process {

    # Prepare the command to run

    $pinfo = New-Object System.Diagnostics.ProcessStartInfo
    $pinfo.FileName               = "$JAVA_PATH"
    $pinfo.Arguments              = "-version"
    $pinfo.UseShellExecute        = $false
    $pinfo.RedirectStandardError  = $true

    # Run the process and capture output

    $p = New-Object System.Diagnostics.Process
    $p.StartInfo = $pinfo
    $p.Start() > $null
    $p.WaitForExit()

    # Extract version number from the output

    $Version  = $( $p.StandardError.ReadToEnd() ).Split( [Environment]::NewLine )
    $Version  = [Regex]::Match( $Version[0], '(\d+).(\d+).(\d+)' ).Value
  }
  end {

    $Name     = "Java"
    $Cur      = "$Version"
    $Req      = "$REQUIRED_JAVA"
    $Params   = @( $Name, $Cur, $Req )

    $Err = Compare-Versions ( $Params )
    if ( $Err ) {
      $Msg = $Msg[0] -f $NAME, $Params[1], $Params[2]
      [Console]::Error.WriteLine( $Msg )
      exit 1
    }
  }
}

function Run-Validations {
  Search-Java-Bin > $null
  Check-JVM-Version
}

function Parse-Arguments {

  # Read parameters to extract the MULE_HOMEs

  for ( $i = 0; $i -lt $CmdArgs.Count; $i++ ) {
    switch -Regex ( $CmdArgs[$i] ) {

      "(rollback|upgrade)" {
        $ACTION = "$_"
        break
      }

      "^(-o|-o=*|--old-mule-distro|--old-mule-distro=*)" {
        if ( $_ -match "=" ) {
          $Param = $_
          $Param = $Param.Split("=")[1]
          $MULE_HOME_OUTDATED = $Param
        }
        else {
          $i++
          $MULE_HOME_OUTDATED = $CmdArgs[$i]
        }
        break
      }

      "^(-n|-n=*|--new-mule-distro|--new-mule-distro=*)" {
        if ( $_ -match "=" ) {
          $Param = $_
          $Param = $Param.Split("=")[1]
          $MULE_HOME_UPDATED = $Param
        }
        else {
          $i++
          $MULE_HOME_UPDATED = $CmdArgs[$i]
        }
        break
      }
    }
  }
}

function Test-Homes {

  # Check that MULE_HOMEs are readable

  foreach ( $Path in "$MULE_HOME_OUTDATED", "$MULE_HOME_UPDATED" ) {

    # If it's not empty
    if ( $Path ) {

      # If it's not a directory
      if ( ! ( Test-Path -Path $Path -PathType Container ) ) {
        [Console]::Error.WriteLine( $Msg[7] )
        exit 1
      }
    }
  }
}

function Restore-Agent {

  # After upgrading, mule-agent-plugin doesn't exist, so If that's empty, grab a copy from the backup

  $Src  = "$MULE_HOME_OUTDATED/.backup/*plugins/mule-agent-plugin" | Resolve-Path 2> $null
  $Dst  = "$MULE_HOME_OUTDATED/*plugins" | Resolve-Path 2> $null

  if ( $Src ) {
    $ValidSrc = Test-Path -Path "$Src"
  }

  if ( $Dst ) {
    $Obj = $Dst.Path + "/mule-agent-plugin"
    $ValidDst = Test-Path -Path "$Obj"
  }

  if ( $ValidSrc -and -not $ValidDst ) {
    $f = $Msg[10] -f $Dst
    [Console]::Out.WriteLine( $f )

    Copy-Item -Path $Src -Destination $Dst -Recurse -Force
  }
}

function Agent-Already-Exist {

  $Upgrade = $true

  $Obj = $ObjExist = $null
  $Obj = "$MULE_HOME_OUTDATED/*plugins/mule-agent-plugin/lib/mule-agent-api-*.*.*.jar" | Resolve-Path 2> $null
  if ( $Obj.Path ) {
    $ObjExist = Test-Path -Path $Obj.Path -PathType Leaf
    if ( ! $ObjExist ) {
      $Upgrade = $false
    }
    else {
      $OUTDATED_JAR = $Obj.Path
    }
  }
  else {
    [Console]::Out.WriteLine( $Msg[11] )
    $Upgrade = $false
  }

  $Obj = $ObjExist = $null
  $Obj = "$MULE_HOME_UPDATED/tools/agent-setup-*.*.*-amc-final.jar" | Resolve-Path 2> $null
  if ( $Obj.Path ) {
    $ObjExist = Test-Path -Path $Obj.Path -PathType Leaf
    if ( ! $ObjExist ) {
      $Upgrade = $false
    }
    else {
      $UPDATED_JAR = $Obj.Path
    }
  }
  else {
    [Console]::Out.WriteLine( $Msg[12] )
    $Upgrade = $false
  }

  return $Upgrade
}

function New-Agent-Version-Available {

  $OldVer = "$OUTDATED_JAR".Split("/")[-1]
  $OldVer = [Regex]::Match( $OldVer, '(\d+).(\d+).(\d+)' ).Value

  $NewVer = "$UPDATED_JAR".Split("/")[-1]
  $NewVer = [Regex]::Match( $NewVer, '(\d+).(\d+).(\d+)' ).Value

  $Name   = "Mule Agent"
  $Cur    = "$OldVer"
  $Req    = "$NewVer"
  $Params = @( $Name, $Cur, $Req )

  # Export info to the global variables:

	$OLD_AGENT_VERSION="$OldVer"
	$NEW_AGENT_VERSION="$NewVer"

  return Compare-Versions ( $Params )
}

function Upgrade-Agent {

  # These are the files we need to migrate manually
  $Files = @(
    "bin/amc_setup*"
    "tools/agent-setup*"
  )

  [Console]::Out.WriteLine( $Msg[3] )

  $f = $Msg[4] -f $OLD_AGENT_VERSION, $NEW_AGENT_VERSION
  [Console]::Out.WriteLine( $f )

  # Copy individual files manually

  Copy-Item -Path $MULE_HOME_UPDATED/$Files[0] -Recurse $MULE_HOME_OUTDATED/bin/
  Copy-Item -Path $MULE_HOME_UPDATED/$Files[1] -Recurse $MULE_HOME_OUTDATED/tools/

  # Run the upgrade, output is grayed out

  $JAVA_ARGS = @( $Env:JAVA_OPTS, "-jar", $UPDATED_JAR, "--update", "--mule-home", $MULE_HOME_OUTDATED)

  $pinfo = New-Object System.Diagnostics.ProcessStartInfo -Prop @{
    FileName                  = "$JAVA_PATH"
    Arguments                 = $JAVA_ARGS
    UseShellExecute           = $false
    RedirectStandardOutput    = $false  # <--- Should be True, but that makes the script to hang, weird
    RedirectStandardError     = $false
  }

  [Console]::ForegroundColor = [System.ConsoleColor]::DarkGray
  $p = New-Object System.Diagnostics.Process
  $p.StartInfo = $pinfo
  $p.Start() > $null
  $p.WaitForExit()
  [Console]::ResetColor()

  # Upgrade completed
  $f = $Msg[5] -f $p.ExitCode
  [Console]::Out.WriteLine( $f )
}

function Setup-Mule-Agent {

  # If the Mule Agent is installed in the current Mule distro, check
	# if the new distro ships a newer version of the Mule Agent
	# If it does, update current Mule Agent by taking it from the newer
	# distro

  if ( $ACTION -eq "upgrade" ) {

    # When upgrading, we need both -o and -n

    if ( ! $MULE_HOME_OUTDATED -or ! $MULE_HOME_UPDATED ) {
      # No action taken
      return
    }

    New-Separator ( $Msg[8] )
    Restore-Agent
    if ( Agent-Already-Exist ) {
      if ( New-Agent-Version-Available ) {
        Upgrade-Agent
      }
      else {
        $f = $Msg[6] -f $OLD_AGENT_VERSION
        [Console]::Error.WriteLine( $f )
      }
    }
    else {
      return
    }
  }
}

function Run-Program {

  New-Separator ( $Msg[9] )

  # If there are whitespaces, put them into an array

  if ( $CmdArgs -match "\s" ) {
    $CmdArgs = $CmdArgs.Split( "\s" )
  }

  $JAVA_ARGS = @( $Env:JAVA_OPTS, "-jar", ( Resolve-Path $JAR_PATTERN ) ) + $CmdArgs

  $pinfo = New-Object System.Diagnostics.ProcessStartInfo -Prop @{
    FileName            = "$JAVA_PATH"
    Arguments           = $JAVA_ARGS
    UseShellExecute     = $false
    }

  $p = New-Object System.Diagnostics.Process
  $p.StartInfo = $pinfo
  $p.Start() > $null
  $p.WaitForExit()
}

function Main {
  Run-Validations
  Parse-Arguments
  Test-Homes
  Run-Program
  Setup-Mule-Agent
}

$CmdArgs = $args
Main
